@@ -8,13 +8,10 @@ from io import BufferedIOBase, StringIO, BytesIO
 from itertools import chain
 from uuid import uuid4
 
-import boto.s3.connection
-import boto.s3.multipart
-import boto.gs.connection
-import boto.s3.key
-import boto.gs.key
+import botocore.config
+import botocore.exceptions
+import boto3.session
 
-from boto.exception import S3ResponseError
 from botocore.signers import CloudFrontSigner
 from cachetools.func import lru_cache
 from cryptography.hazmat.backends import default_backend
@@ -40,13 +37,52 @@ multipart_uploads_completed = Counter(
 )
 
 
+_PartUpload = namedtuple("_PartUpload", ["part_number", "e_tag"])
 _PartUploadMetadata = namedtuple("_PartUploadMetadata", ["path", "offset", "length"])
 _CHUNKS_KEY = "chunks"
 
 
+# This is for HEAD requests to check if a key exists.
+# Since the HEAD request does not have a response body, boto3 uses the status code as error code
+# Ref: https://github.com/boto/boto3/issues/2442
+_MISSING_KEY_ERROR_CODES = ("NoSuchKey", "404")
+_LIST_OBJECT_VERSIONS = {"v1": "list_objects", "v2": "list_objects_v2"}
+
+
+def _build_endpoint_url(hostname, port=None, is_secure=True):
+    """Normalize the formats from boto2 and boto3. """
+
+    # If the scheme is not in the hostname, check if is_secure is set to set http or https as the scheme
+    if not hostname.startswith("http://") and not hostname.startswith("https://"):
+        hostname = "https://" + hostname if is_secure else "http://" + hostname
+
+    # If port is set but not already part of the hostname
+    if port and not hostname.split(":")[-1].strip(" /").isdigit():
+        hostname = hostname + ":" + str(port)
+
+    return hostname
+
+
 class StreamReadKeyAsFile(BufferedIOBase):
+    """
+    Wrapper for botocore.StreamingBody
+
+    NOTE: Moving from boto2 to boto3, the file-like interface implementation between bn both versions behave differently.
+          In boto3, once the stream is out of data, it will keep returning an empty string until it is closed.
+          In boto2, once the stream is out of data, it will return an empty string once, and subsequent read will start reading
+          the stream from the beginning again (by making a new GET request, since it is server streaming the content
+          (calling `close()` does not seem to have any effect). This wrapper provides a more consistent file-like interface and
+          allows to check whether a key is closed or not.
+
+          References:
+          - https://github.com/boto/boto/blob/develop/boto/s3/key.py#L51 (boto2's "Key")
+          - https://github.com/boto/botocore/blob/develop/botocore/response.py (boto3's "StreamingBody")
+          - https://github.com/boto/boto3/issues/564
+    """
+
     def __init__(self, key):
         self._key = key
+        self._closed = False
 
     def read(self, amt=None):
         if self.closed:
@@ -60,10 +96,11 @@ class StreamReadKeyAsFile(BufferedIOBase):
 
     @property
     def closed(self):
-        return self._key.closed
+        return self._closed
 
     def close(self):
-        self._key.close(fast=True)
+        self._closed = True
+        self._key.close()
 
 
 class _CloudStorage(BaseStorageV2):
@@ -71,7 +108,6 @@ class _CloudStorage(BaseStorageV2):
         self,
         context,
         connection_class,
-        key_class,
         connect_kwargs,
         upload_params,
         storage_path,
@@ -90,35 +126,69 @@ class _CloudStorage(BaseStorageV2):
         self._secret_key = secret_key
         self._root_path = storage_path
         self._connection_class = connection_class
-        self._key_class = key_class
         self._upload_params = upload_params
         self._connect_kwargs = connect_kwargs
         self._cloud_conn = None
         self._cloud_bucket = None
         self._context = context
+        self._list_object_version = _LIST_OBJECT_VERSIONS["v2"]
+
+        self._session = self._connection_class(
+            aws_access_key_id=self._access_key,
+            aws_secret_access_key=self._secret_key,
+        )
 
     def _initialize_cloud_conn(self):
         if not self._initialized:
-            self._cloud_conn = self._connection_class(
-                self._access_key, self._secret_key, **self._connect_kwargs
+            # Low-level client. Needed to generate presigned urls
+            self._cloud_conn = self._session.client("s3", **self._connect_kwargs)
+            self._cloud_bucket = self._session.resource("s3", **self._connect_kwargs).Bucket(
+                self._bucket_name
             )
-            self._cloud_bucket = self._cloud_conn.get_bucket(self._bucket_name, validate=False)
+            # This will raise a ClientError if the bucket does ot exists.
+            # We actually want an exception raised if the bucket does not exists (same as in boto2)
+            self._cloud_conn.head_bucket(Bucket=self._bucket_name)
             self._initialized = True
 
-    def _debug_key(self, key):
+    def _debug_key(self, obj):
         """
         Used for debugging only.
         """
-        orig_meth = key.bucket.connection.make_request
+        import types
 
-        def new_meth(*args, **kwargs):
-            print("#" * 16)
-            print(args)
-            print(kwargs)
-            print("#" * 16)
-            return orig_meth(*args, **kwargs)
+        valid_debug_methods = [
+            "copy",
+            "copy_from",
+            "delete",
+            "download_file",
+            "download_fileobj",
+            "get",
+            "get_available_subresources",
+            "initiate_multipart_upload",
+            "load",
+            "put",
+            "reload",
+            "restore_object",
+            "upload_file",
+            "upload_fileobj",
+        ]
 
-        key.bucket.connection.make_request = new_meth
+        # Wraps the method to debug
+        def debug_method_decorator(f):
+            @wraps(f)
+            def wrapper(self, *args, **kwargs):
+                print("#" * 16)
+                print(args)
+                print(kwargs)
+                print("#" * 16)
+                return f(*args, **kwargs)
+
+            return wrapper
+
+        # Binds the new methods to the instance
+        for method in valid_debug_methods:
+            new_meth = debug_method_decorator(getattr(obj, method))
+            obj.__setattr__(method, types.MethodType(new_meth, obj))
 
     def _init_path(self, path=None):
         path = os.path.join(self._root_path, path) if path else self._root_path
@@ -136,13 +206,13 @@ class _CloudStorage(BaseStorageV2):
     def get_content(self, path):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
+        obj = self.get_cloud_bucket().Object(path)
         try:
-            return key.get_contents_as_string()
-        except S3ResponseError as s3r:
+            return obj.get()["Body"].read()
+        except botocore.exceptions.ClientError as s3r:
             # Raise an IOError in case the key was not found, to maintain the current
             # interface.
-            if s3r.error_code == "NoSuchKey":
+            if s3r.response["Error"]["Code"] in _MISSING_KEY_ERROR_CODES:
                 raise IOError("No such key: '{0}'".format(path))
 
             raise
@@ -150,8 +220,8 @@ class _CloudStorage(BaseStorageV2):
     def put_content(self, path, content):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
-        key.set_contents_from_string(content, **self._upload_params)
+        obj = self.get_cloud_bucket().Object(path)
+        obj.put(Body=content, **self._upload_params)
         return path
 
     def get_supports_resumable_downloads(self):
@@ -162,55 +232,78 @@ class _CloudStorage(BaseStorageV2):
     ):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        k = self._key_class(self._cloud_bucket, path)
+
+        client_method = "get_object"
         if head:
-            return k.generate_url(expires_in, "HEAD")
-        return k.generate_url(expires_in)
+            client_method = "head_object"
+
+        return self.get_cloud_conn().generate_presigned_url(
+            client_method,
+            Params={"Bucket": self._bucket_name, "Key": path},
+            ExpiresIn=expires_in,
+        )
 
     def get_direct_upload_url(self, path, mime_type, requires_cors=True):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
-        url = key.generate_url(300, "PUT", headers={"Content-Type": mime_type}, encrypt_key=True)
-        return url
+        return self.get_cloud_conn().generate_presigned_url(
+            "put_object",
+            Params={
+                "Bucket": self._bucket_name,
+                "Key": path,
+                "ServerSideEncryption": "AES256",
+                "ContentType": mime_type,
+            },
+            ExpiresIn=300,
+        )
 
     def stream_read(self, path):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
-        if not key.exists():
-            raise IOError("No such key: '{0}'".format(path))
+        obj = self.get_cloud_bucket().Object(path)
+        try:
+            obj.load()
+        except botocore.exceptions.ClientError as s3r:
+            if s3r.response["Error"]["Code"] in _MISSING_KEY_ERROR_CODES:
+                raise IOError("No such key: '{0}'".format(path))
+            raise
+
+        buf = obj.get()["Body"]
         while True:
-            buf = key.read(self.buffer_size)
-            if not buf:
+            data = buf.read(self.buffer_size)
+            if not data:
                 break
-            yield buf
+            yield data
 
     def stream_read_file(self, path):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
-        if not key.exists():
-            raise IOError("No such key: '{0}'".format(path))
-        return StreamReadKeyAsFile(key)
+        obj = self.get_cloud_bucket().Object(path)
+        try:
+            obj.load()
+        except botocore.exceptions.ClientError as s3r:
+            if s3r.response["Error"]["Code"] in _MISSING_KEY_ERROR_CODES:
+                raise IOError("No such key: '{0}'".format(path))
+            raise
+        return StreamReadKeyAsFile(obj.get()["Body"])
 
     def __initiate_multipart_upload(self, path, content_type, content_encoding):
         # Minimum size of upload part size on S3 is 5MB
         self._initialize_cloud_conn()
         path = self._init_path(path)
+        obj = self.get_cloud_bucket().Object(path)
 
         metadata = {}
         if content_type is not None:
-            metadata["Content-Type"] = content_type
+            metadata["ContentType"] = content_type
 
         if content_encoding is not None:
-            metadata["Content-Encoding"] = content_encoding
+            metadata["ContentEncoding"] = content_encoding
+
+        metadata = {**metadata, **self._upload_params}
 
         multipart_uploads_started.inc()
-
-        return self._cloud_bucket.initiate_multipart_upload(
-            path, metadata=metadata, **self._upload_params
-        )
+        return obj.initiate_multipart_upload(**metadata)
 
     def stream_write(self, path, fp, content_type=None, content_encoding=None):
         """
@@ -243,12 +336,12 @@ class _CloudStorage(BaseStorageV2):
 
         try:
             mp = self.__initiate_multipart_upload(path, content_type, content_encoding)
-        except S3ResponseError as e:
+        except botocore.exceptions.ClientError as s3r:
             logger.exception("Exception when initiating multipart upload")
-            return 0, e
+            return 0, s3r
 
-        # We are going to reuse this but be VERY careful to only read the number of bytes written to it
-        buf = BytesIO()
+        # [_PartUpload]
+        upload_parts = []
 
         num_part = 1
         total_bytes_written = 0
@@ -258,71 +351,103 @@ class _CloudStorage(BaseStorageV2):
                 # We never want to ask for more bytes than our caller has indicated to copy
                 bytes_to_copy = min(bytes_to_copy, size - total_bytes_written)
 
-            buf.seek(0)
-            try:
-                # Stage the bytes into the buffer for use with the multipart upload file API
-                bytes_staged = self.stream_write_to_fp(fp, buf, bytes_to_copy)
-                if bytes_staged == 0:
-                    break
+            with BytesIO() as buf:
+                try:
+                    # Stage the bytes into the buffer for use with the multipart upload file API
+                    bytes_staged = self.stream_write_to_fp(fp, buf, bytes_to_copy)
 
-                buf.seek(0)
-                mp.upload_part_from_file(buf, num_part, size=bytes_staged)
-                total_bytes_written += bytes_staged
-                num_part += 1
-            except (S3ResponseError, IOError) as e:
-                logger.warn(
-                    "Error when writing to stream in stream_write_internal at path %s: %s", path, e
-                )
-                write_error = e
+                    if bytes_staged == 0:
+                        break
 
-                multipart_uploads_completed.inc()
+                    buf.seek(0)
+                    part = mp.Part(num_part)
+                    part_upload = part.upload(
+                        Body=buf,
+                        ContentLength=bytes_staged,
+                    )
+                    upload_parts.append(_PartUpload(num_part, part_upload["ETag"]))
+                    total_bytes_written += bytes_staged
+                    num_part += 1
+                except (
+                    botocore.exceptions.ClientError,
+                    botocore.exceptions.ConnectionClosedError,
+                    IOError,
+                ) as e:
+                    logger.warn(
+                        "Error when writing to stream in stream_write_internal at path %s: %s",
+                        path,
+                        e,
+                    )
+                    write_error = e
 
-                if cancel_on_error:
-                    try:
-                        mp.cancel_upload()
-                    except (S3ResponseError, IOError):
-                        logger.exception("Could not cancel upload")
+                    multipart_uploads_completed.inc()
 
-                    return 0, write_error
-                else:
-                    break
+                    if cancel_on_error:
+                        try:
+                            mp.abort()
+                        except (botocore.exceptions.ClientError, IOError):
+                            logger.exception("Could not cancel upload")
+
+                        return 0, write_error
+                    else:
+                        break
 
         if total_bytes_written > 0:
             multipart_uploads_completed.inc()
-
-            self._perform_action_with_retry(mp.complete_upload)
+            self._perform_action_with_retry(
+                mp.complete,
+                MultipartUpload={
+                    "Parts": [{"ETag": p.e_tag, "PartNumber": p.part_number} for p in upload_parts],
+                },
+            )
 
         return total_bytes_written, write_error
 
     def exists(self, path):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
-        return key.exists()
+        obj = self.get_cloud_bucket().Object(path)
+        try:
+            obj.load()
+        except botocore.exceptions.ClientError as s3r:
+            if s3r.response["Error"]["Code"] in _MISSING_KEY_ERROR_CODES:
+                return False
+            raise
+        return True
 
     def remove(self, path):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
-        if key.exists():
-            # It's a file
-            key.delete()
+        obj = self.get_cloud_bucket().Object(path)
+        try:
+            obj.load()
+            obj.delete()
             return
+        except botocore.exceptions.ClientError as s3r:
+            if not s3r.response["Error"]["Code"] in _MISSING_KEY_ERROR_CODES:
+                raise
         # We assume it's a directory
         if not path.endswith("/"):
             path += "/"
-        for key in self._cloud_bucket.list(prefix=path):
-            key.delete()
+
+        paginator = self.get_cloud_conn().get_paginator(self._list_object_version)
+        for page in paginator.paginate(Bucket=self._bucket_name, Prefix=path):
+            for content in page.get("Contents", ()):
+                obj = self.get_cloud_bucket().Object(content["Key"])
+                obj.delete()
 
     def get_checksum(self, path):
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
-        k = self._cloud_bucket.lookup(key)
-        if k is None:
-            raise IOError("No such key: '{0}'".format(path))
+        obj = self.get_cloud_bucket().Object(path)
+        try:
+            obj.load()
+        except botocore.exceptions.ClientError as s3r:
+            if s3r.response["Error"]["Code"] in _MISSING_KEY_ERROR_CODES:
+                raise IOError("No such key: '{0}'".format(path))
+            raise
 
-        return k.etag[1:-1][:7]
+        return obj.e_tag[1:-1][:7]
 
     def copy_to(self, destination, path):
         """
@@ -367,10 +492,11 @@ class _CloudStorage(BaseStorageV2):
             )
 
             source_path = self._init_path(path)
-            source_key = self._key_class(self._cloud_bucket, source_path)
+            source_obj = self.get_cloud_bucket().Object(source_path)
 
             dest_path = destination._init_path(path)
-            source_key.copy(destination._cloud_bucket, dest_path)
+            dest_obj = destination.get_cloud_bucket().Object(dest_path)
+            dest_obj.copy_from(CopySource={"Bucket": source_obj.bucket_name, "Key": source_obj.key})
             return
 
         # Fallback to a slower, default copy.
@@ -425,12 +551,18 @@ class _CloudStorage(BaseStorageV2):
             abs_final_path = self._init_path(final_path)
 
             # Let the copy raise an exception if it fails.
-            self._cloud_bucket.copy_key(abs_final_path, self._bucket_name, chunk_path)
+            #
+            # TODO(kleesc): copy_from() is used instead of copy, since the latter is a managed transfer which uses S3's
+            #               multipart api, which GCS and Rados does not support. Going forward, we should try moving
+            #               non-aws implementations to use library's with better support (e.g GCS supports its own version
+            #               of parallel uploads).
+            new_obj = self.get_cloud_bucket().Object(abs_final_path)
+            new_obj.copy_from(CopySource={"Bucket": self._bucket_name, "Key": chunk_path})
 
             # Attempt to clean up the old chunk.
             try:
-                self._cloud_bucket.delete_key(chunk_path)
-            except IOError:
+                self.get_cloud_bucket().Object(chunk_path).delete()
+            except (botocore.exceptions.ClientError, IOError):
                 # We failed to delete a chunk. This sucks, but we shouldn't fail the push.
                 msg = "Failed to clean up chunk %s for move of %s"
                 logger.exception(msg, chunk_path, abs_final_path)
@@ -442,8 +574,8 @@ class _CloudStorage(BaseStorageV2):
             # Attempt to clean up all the chunks.
             for chunk in chunk_list:
                 try:
-                    self._cloud_bucket.delete_key(self._init_path(chunk.path))
-                except IOError:
+                    self.get_cloud_bucket().Object(chunk.path).delete()
+                except (botocore.exceptions.ClientError, IOError):
                     # We failed to delete a chunk. This sucks, but we shouldn't fail the push.
                     msg = "Failed to clean up chunk %s for reupload of %s"
                     logger.exception(msg, chunk.path, final_path)
@@ -454,10 +586,14 @@ class _CloudStorage(BaseStorageV2):
         # an action. The recommendation is to simply try calling the action again.
         for remaining_retries in range(2, -1, -1):
             try:
-                action(*args, **kwargs)
+                return action(*args, **kwargs)
                 break
-            except S3ResponseError as s3re:
-                if remaining_retries and s3re.status == 200 and s3re.error_code == "InternalError":
+            except botocore.exceptions.ClientError as s3re:
+                if (
+                    remaining_retries
+                    and s3re.response["Error"]["HTTPStatusCode"] == 200
+                    and s3re.response["Error"]["Code"] == "InternalError"
+                ):
                     # Weird internal error case. Retry.
                     continue
 
@@ -511,23 +647,40 @@ class _CloudStorage(BaseStorageV2):
                     [_CloudStorage._rechunk(c, self.maximum_chunk_size) for c in chunk_list]
                 )
 
+                # [_PartUpload]
+                upload_parts = []
                 for index, chunk in enumerate(updated_chunks):
                     abs_chunk_path = self._init_path(chunk.path)
-                    self._perform_action_with_retry(
-                        mpu.copy_part_from_key,
-                        self.get_cloud_bucket().name,
-                        abs_chunk_path,
-                        index + 1,
-                        start=chunk.offset,
-                        end=chunk.length + chunk.offset - 1,
+
+                    part = mpu.Part(index + 1)
+                    part_copy = part.copy_from(
+                        CopySource={"Bucket": self.get_cloud_bucket().name, "Key": abs_chunk_path},
+                        CopySourceRange="bytes=%s-%s"
+                        % (chunk.offset, chunk.length + chunk.offset - 1),
                     )
 
-                self._perform_action_with_retry(mpu.complete_upload)
-            except IOError as ioe:
+                    part_copy = self._perform_action_with_retry(
+                        mpu.Part(index + 1).copy_from,
+                        CopySource={"Bucket": self.get_cloud_bucket().name, "Key": abs_chunk_path},
+                        CopySourceRange="bytes=%s-%s"
+                        % (chunk.offset, chunk.length + chunk.offset - 1),
+                    )
+
+                    upload_parts.append(_PartUpload(index + 1, part_copy["CopyPartResult"]["ETag"]))
+
+                self._perform_action_with_retry(
+                    mpu.complete,
+                    MultipartUpload={
+                        "Parts": [
+                            {"ETag": p.e_tag, "PartNumber": p.part_number} for p in upload_parts
+                        ]
+                    },
+                )
+            except (botocore.exceptions.ClientError, IOError) as ioe:
                 # Something bad happened, log it and then give up
                 msg = "Exception when attempting server-side assembly for: %s"
                 logger.exception(msg, final_path)
-                mpu.cancel_upload()
+                mpu.abort()
                 raise ioe
 
         else:
@@ -551,26 +704,22 @@ class S3Storage(_CloudStorage):
         s3_bucket,
         s3_access_key=None,
         s3_secret_key=None,
+        # Boto2 backward compatible options (host excluding scheme or port)
         host=None,
         port=None,
+        # Boto3 options (Full url including scheme anbd optionally port)
+        endpoint_url=None,
     ):
-        upload_params = {
-            "encrypt_key": True,
-        }
+        upload_params = {"ServerSideEncryption": "AES256"}
         connect_kwargs = {}
-        if host:
-            if host.startswith("http:") or host.startswith("https:"):
-                raise ValueError("host name must not start with http:// or https://")
-
-            connect_kwargs["host"] = host
-
-        if port:
-            connect_kwargs["port"] = int(port)
+        if host or endpoint_url:
+            connect_kwargs["endpoint_url"] = (
+                endpoint_url or _build_endpoint_url(host, port=port, is_secure=True),
+            )
 
         super(S3Storage, self).__init__(
             context,
-            boto.s3.connection.S3Connection,
-            boto.s3.key.Key,
+            boto3.session.Session,
             connect_kwargs,
             upload_params,
             storage_path,
@@ -582,35 +731,37 @@ class S3Storage(_CloudStorage):
         self.maximum_chunk_size = 5 * 1024 * 1024 * 1024  # 5GB.
 
     def setup(self):
-        self.get_cloud_bucket().set_cors_xml(
-            """<?xml version="1.0" encoding="UTF-8"?>
-      <CORSConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
-          <CORSRule>
-              <AllowedOrigin>*</AllowedOrigin>
-              <AllowedMethod>GET</AllowedMethod>
-              <MaxAgeSeconds>3000</MaxAgeSeconds>
-              <AllowedHeader>Authorization</AllowedHeader>
-          </CORSRule>
-          <CORSRule>
-              <AllowedOrigin>*</AllowedOrigin>
-              <AllowedMethod>PUT</AllowedMethod>
-              <MaxAgeSeconds>3000</MaxAgeSeconds>
-              <AllowedHeader>Content-Type</AllowedHeader>
-              <AllowedHeader>x-amz-acl</AllowedHeader>
-              <AllowedHeader>origin</AllowedHeader>
-          </CORSRule>
-      </CORSConfiguration>"""
+        self.get_cloud_bucket().Cors().put(
+            CORSConfiguration={
+                "CORSRules": [
+                    {
+                        "AllowedOrigins": ["*"],
+                        "AllowedMethods": ["GET"],
+                        "MaxAgeSeconds": 3000,
+                        "AllowedHeaders": ["Authorization"],
+                    },
+                    {
+                        "AllowedOrigins": ["*"],
+                        "AllowedMethods": ["PUT"],
+                        "MaxAgeSeconds": 3000,
+                        "AllowedHeaders": ["Content-Type", "x-amz-acl", "origin"],
+                    },
+                ]
+            }
         )
 
 
 class GoogleCloudStorage(_CloudStorage):
+    ENDPOINT_URL = "https://storage.googleapis.com"
+
     def __init__(self, context, storage_path, access_key, secret_key, bucket_name):
+        # GCS does not support ListObjectV2
+        self._list_object_version = _LIST_OBJECT_VERSIONS["v1"]
         upload_params = {}
-        connect_kwargs = {}
+        connect_kwargs = {"endpoint_url": GoogleCloudStorage.ENDPOINT_URL}
         super(GoogleCloudStorage, self).__init__(
             context,
-            boto.gs.connection.GSConnection,
-            boto.gs.key.Key,
+            boto3.session.Session,
             connect_kwargs,
             upload_params,
             storage_path,
@@ -619,24 +770,55 @@ class GoogleCloudStorage(_CloudStorage):
             secret_key,
         )
 
+        # Workaround for setting GCS cors at runtime with boto
+        cors_xml = """<?xml version="1.0" encoding="UTF-8"?>
+        <CorsConfig>
+          <Cors>
+            <Origins>
+              <Origin>*</Origin>
+            </Origins>
+            <Methods>
+              <Method>GET</Method>
+              <Method>PUT</Method>
+            </Methods>
+            <ResponseHeaders>
+              <ResponseHeader>Content-Type</ResponseHeader>
+            </ResponseHeaders>
+            <MaxAgeSec>3000</MaxAgeSec>
+          </Cors>
+        </CorsConfig>"""
+
+        def update_cors_xml(request, **kwargs):
+            request.data = cors_xml
+
+        def validate_cors_xml(request, **kwargs):
+            assert request.body == cors_xml
+
+        self._session.events.register("request-created.s3.PutBucketCors", update_cors_xml)
+        self._session.events.register("before-send.s3.PutBucketCors", validate_cors_xml)
+
     def setup(self):
-        self.get_cloud_bucket().set_cors_xml(
-            """<?xml version="1.0" encoding="UTF-8"?>
-      <CorsConfig>
-        <Cors>
-          <Origins>
-            <Origin>*</Origin>
-          </Origins>
-          <Methods>
-            <Method>GET</Method>
-            <Method>PUT</Method>
-          </Methods>
-          <ResponseHeaders>
-            <ResponseHeader>Content-Type</ResponseHeader>
-          </ResponseHeaders>
-          <MaxAgeSec>3000</MaxAgeSec>
-        </Cors>
-      </CorsConfig>"""
+        self._initialize_cloud_conn()
+        self.get_cloud_bucket().Cors().put(
+            CORSConfiguration={
+                "CORSRules": [
+                    {
+                        "AllowedOrigins": ["*"],
+                        "AllowedMethods": ["GET", "PUT"],
+                        "MaxAgeSeconds": 3000,
+                        "AllowedHeaders": ["Content-Type"],
+                    },
+                ]
+            }
+        )
+
+    def get_direct_download_url(
+        self, path, request_ip=None, expires_in=60, requires_cors=False, head=False
+    ):
+        return (
+            super(GoogleCloudStorage, self)
+            .get_direct_download_url(path, request_ip, expires_in, requires_cors, head)
+            .replace("AWSAccessKeyId", "GoogleAccessId")
         )
 
     def _stream_write_internal(
@@ -658,24 +840,30 @@ class GoogleCloudStorage(_CloudStorage):
         # Minimum size of upload part size on S3 is 5MB
         self._initialize_cloud_conn()
         path = self._init_path(path)
-        key = self._key_class(self._cloud_bucket, path)
+        obj = self.get_cloud_bucket().Object(path)
 
+        extra_args = {}
         if content_type is not None:
-            key.set_metadata("Content-Type", content_type)
+            extra_args["ContentType"] = content_type
 
         if content_encoding is not None:
-            key.set_metadata("Content-Encoding", content_encoding)
+            extra_args["ContentEncoding"] = content_encoding
 
         if size != filelike.READ_UNTIL_END:
             fp = filelike.StreamSlice(fp, 0, size)
 
-        # TODO figure out how to handle cancel_on_error=False
-        try:
-            key.set_contents_from_stream(fp)
-        except IOError as ex:
-            return 0, ex
+        with BytesIO() as buf:
+            # Stage the bytes into the buffer for use with the multipart upload file API
+            bytes_staged = self.stream_write_to_fp(fp, buf, size)
+            buf.seek(0)
 
-        return key.size, None
+            # TODO figure out how to handle cancel_on_error=False
+            try:
+                obj.put(Body=buf, **extra_args)
+            except Exception as ex:
+                return 0, ex
+
+        return bytes_staged, None
 
     def complete_chunked_upload(self, uuid, final_path, storage_metadata):
         self._initialize_cloud_conn()
@@ -701,18 +889,12 @@ class RadosGWStorage(_CloudStorage):
     ):
         upload_params = {}
         connect_kwargs = {
-            "host": hostname,
-            "is_secure": is_secure,
-            "calling_format": boto.s3.connection.OrdinaryCallingFormat(),
+            "endpoint_url": _build_endpoint_url(hostname, port=port, is_secure=is_secure),
         }
 
-        if port:
-            connect_kwargs["port"] = int(port)
-
         super(RadosGWStorage, self).__init__(
             context,
-            boto.s3.connection.S3Connection,
-            boto.s3.key.Key,
+            boto3.session.Session,
             connect_kwargs,
             upload_params,
             storage_path,
